@@ -1,6 +1,13 @@
 //! This crate provides a fully written in Rust memory allocator for Vulkan and DirectX 12.
 //!
-//! ## Setting up the Vulkan memory allocator
+//! # [Windows-rs] and [winapi]
+//!
+//! `gpu-allocator` recently migrated from [winapi] to [windows-rs] but still provides convenient helpers to convert to and from [winapi] types, enabled when compiling with the `public-winapi` crate feature.
+//!
+//! [Windows-rs]: https://github.com/microsoft/windows-rs
+//! [winapi]: https://github.com/retep998/winapi-rs
+//!
+//! # Setting up the Vulkan memory allocator
 //!
 //! ```no_run
 //! # #[cfg(feature = "vulkan")]
@@ -17,13 +24,14 @@
 //!     physical_device,
 //!     debug_settings: Default::default(),
 //!     buffer_device_address: true,  // Ideally, check the BufferDeviceAddressFeatures struct.
+//!     allocation_sizes: Default::default(),
 //! });
 //! # }
 //! # #[cfg(not(feature = "vulkan"))]
 //! # fn main() {}
 //! ```
 //!
-//! ## Simple Vulkan allocation example
+//! # Simple Vulkan allocation example
 //!
 //! ```no_run
 //! # #[cfg(feature = "vulkan")]
@@ -34,13 +42,13 @@
 //! # let device = todo!();
 //! # let instance = todo!();
 //! # let physical_device = todo!();
-//!
 //! # let mut allocator = Allocator::new(&AllocatorCreateDesc {
 //! #     instance,
 //! #     device,
 //! #     physical_device,
 //! #     debug_settings: Default::default(),
 //! #     buffer_device_address: true,  // Ideally, check the BufferDeviceAddressFeatures struct.
+//! #     allocation_sizes: Default::default(),
 //! # }).unwrap();
 //!
 //! // Setup vulkan info
@@ -57,6 +65,7 @@
 //!         requirements,
 //!         location: MemoryLocation::CpuToGpu,
 //!         linear: true, // Buffers are always linear
+//!         allocation_scheme: AllocationScheme::GpuAllocatorManaged,
 //!     }).unwrap();
 //!
 //! // Bind memory to the buffer
@@ -70,83 +79,80 @@
 //! # fn main() {}
 //! ```
 //!
-//! ## Setting up the D3D12 memory allocator
+//! # Setting up the D3D12 memory allocator
 //!
 //! ```no_run
 //! # #[cfg(feature = "d3d12")]
 //! # fn main() {
 //! use gpu_allocator::d3d12::*;
-//! # use winapi::um::d3d12;
 //! # let device = todo!();
 //!
 //! let mut allocator = Allocator::new(&AllocatorCreateDesc {
-//!     device,
+//!     device: ID3D12DeviceVersion::Device(device),
 //!     debug_settings: Default::default(),
+//!     allocation_sizes: Default::default(),
 //! });
 //! # }
 //! # #[cfg(not(feature = "d3d12"))]
 //! # fn main() {}
 //! ```
 //!
-//! ## Simple d3d12 allocation example
+//! # Simple d3d12 allocation example
 //!
 //! ```no_run
-//! # #[cfg(all(feature = "d3d12", feature = "public-winapi"))]
-//! # fn main() {
+//! # #[cfg(feature = "d3d12")]
+//! # fn main() -> windows::core::Result<()> {
 //! use gpu_allocator::d3d12::*;
 //! use gpu_allocator::MemoryLocation;
-//! # use winapi::um::d3d12;
-//! # use winapi::shared::{dxgiformat, dxgitype, winerror};
-//! # let device: *mut d3d12::ID3D12Device = todo!();
+//! # use windows::Win32::Graphics::{Dxgi, Direct3D12};
+//! # let device = todo!();
 //!
 //! # let mut allocator = Allocator::new(&AllocatorCreateDesc {
-//! #     device: Dx12DevicePtr(device as *mut _),
+//! #     device: ID3D12DeviceVersion::Device(device),
 //! #     debug_settings: Default::default(),
+//! #     allocation_sizes: Default::default(),
 //! # }).unwrap();
 //!
-//! let buffer_desc = d3d12::D3D12_RESOURCE_DESC {
-//!     Dimension: d3d12::D3D12_RESOURCE_DIMENSION_BUFFER,
+//! let buffer_desc = Direct3D12::D3D12_RESOURCE_DESC {
+//!     Dimension: Direct3D12::D3D12_RESOURCE_DIMENSION_BUFFER,
 //!     Alignment: 0,
 //!     Width: 512,
 //!     Height: 1,
 //!     DepthOrArraySize: 1,
 //!     MipLevels: 1,
-//!     Format: dxgiformat::DXGI_FORMAT_UNKNOWN,
-//!     SampleDesc: dxgitype::DXGI_SAMPLE_DESC {
+//!     Format: Dxgi::Common::DXGI_FORMAT_UNKNOWN,
+//!     SampleDesc: Dxgi::Common::DXGI_SAMPLE_DESC {
 //!         Count: 1,
 //!         Quality: 0,
 //!     },
-//!     Layout: d3d12::D3D12_TEXTURE_LAYOUT_ROW_MAJOR,
-//!     Flags: d3d12::D3D12_RESOURCE_FLAG_NONE,
+//!     Layout: Direct3D12::D3D12_TEXTURE_LAYOUT_ROW_MAJOR,
+//!     Flags: Direct3D12::D3D12_RESOURCE_FLAG_NONE,
 //! };
 //! let allocation_desc = AllocationCreateDesc::from_d3d12_resource_desc(
-//!     allocator.device(),
+//!     &allocator.device(),
 //!     &buffer_desc,
 //!     "Example allocation",
 //!     MemoryLocation::GpuOnly,
 //! );
 //! let allocation = allocator.allocate(&allocation_desc).unwrap();
-//! let mut resource: *mut d3d12::ID3D12Resource = std::ptr::null_mut();
+//! let mut resource: Option<Direct3D12::ID3D12Resource> = None;
 //! let hr = unsafe {
-//!     device.as_ref().unwrap().CreatePlacedResource(
-//!         allocation.heap().as_winapi_mut(),
+//!     device.CreatePlacedResource(
+//!         allocation.heap(),
 //!         allocation.offset(),
 //!         &buffer_desc,
-//!         d3d12::D3D12_RESOURCE_STATE_COMMON,
-//!         std::ptr::null(),
-//!         &d3d12::IID_ID3D12Resource,
-//!         &mut resource as *mut _ as *mut _,
+//!         Direct3D12::D3D12_RESOURCE_STATE_COMMON,
+//!         None,
+//!         &mut resource,
 //!     )
-//! };
-//! if hr != winerror::S_OK {
-//!     panic!("Failed to create placed resource.");
-//! }
+//! }?;
 //!
 //! // Cleanup
-//! unsafe { resource.as_ref().unwrap().Release() };
+//! drop(resource);
 //! allocator.free(allocation).unwrap();
+//! # Ok(())
 //! # }
-//! # #[cfg(not(all(feature = "d3d12", feature="public-winapi")))]
+//! # #[cfg(not(feature = "d3d12"))]
 //! # fn main() {}
 //! ```
 
@@ -207,3 +213,65 @@ impl Default for AllocatorDebugSettings {
         }
     }
 }
+
+/// The sizes of the memory blocks that the allocator will create.
+///
+/// Useful for tuning the allocator to your application's needs. For example most games will be fine with the default
+/// values, but eg. an app might want to use smaller block sizes to reduce the amount of memory used.
+///
+/// Clamped between 4MB and 256MB, and rounds up to the nearest multiple of 4MB for alignment reasons.
+#[derive(Clone, Copy, Debug)]
+pub struct AllocationSizes {
+    /// The size of the memory blocks that will be created for the GPU only memory type.
+    ///
+    /// Defaults to 256MB.
+    device_memblock_size: u64,
+    /// The size of the memory blocks that will be created for the CPU visible memory types.
+    ///
+    /// Defaults to 64MB.
+    host_memblock_size: u64,
+}
+
+impl AllocationSizes {
+    pub fn new(device_memblock_size: u64, host_memblock_size: u64) -> Self {
+        const FOUR_MB: u64 = 4 * 1024 * 1024;
+        const TWO_HUNDRED_AND_FIFTY_SIX_MB: u64 = 256 * 1024 * 1024;
+
+        let mut device_memblock_size =
+            device_memblock_size.clamp(FOUR_MB, TWO_HUNDRED_AND_FIFTY_SIX_MB);
+        let mut host_memblock_size =
+            host_memblock_size.clamp(FOUR_MB, TWO_HUNDRED_AND_FIFTY_SIX_MB);
+
+        if device_memblock_size % FOUR_MB != 0 {
+            let val = device_memblock_size / FOUR_MB + 1;
+            device_memblock_size = val * FOUR_MB;
+            log::warn!(
+                "Device memory block size must be a multiple of 4MB, clamping to {}MB",
+                device_memblock_size / 1024 / 1024
+            )
+        }
+
+        if host_memblock_size % FOUR_MB != 0 {
+            let val = host_memblock_size / FOUR_MB + 1;
+            host_memblock_size = val * FOUR_MB;
+            log::warn!(
+                "Host memory block size must be a multiple of 4MB, clamping to {}MB",
+                host_memblock_size / 1024 / 1024
+            )
+        }
+
+        Self {
+            device_memblock_size,
+            host_memblock_size,
+        }
+    }
+}
+
+impl Default for AllocationSizes {
+    fn default() -> Self {
+        Self {
+            device_memblock_size: 256 * 1024 * 1024,
+            host_memblock_size: 64 * 1024 * 1024,
+        }
+    }
+}
