@@ -8,10 +8,26 @@ pub use visualizer::AllocatorVisualizer;
 use super::allocator;
 use super::allocator::AllocationType;
 use ash::vk;
-use log::{debug, warn, Level};
+use core::marker::PhantomData;
+use log::{debug, Level};
 use std::fmt;
 
-use crate::{AllocationError, AllocatorDebugSettings, MemoryLocation, Result};
+use crate::{
+    allocator::fmt_bytes, AllocationError, AllocationSizes, AllocatorDebugSettings, MemoryLocation,
+    Result,
+};
+
+#[derive(Copy, Clone, Debug, Eq, PartialEq)]
+pub enum AllocationScheme {
+    /// Perform a dedicated, driver-managed allocation for the given buffer, allowing
+    /// it to perform optimizations on this type of allocation.
+    DedicatedBuffer(vk::Buffer),
+    /// Perform a dedicated, driver-managed allocation for the given image, allowing
+    /// it to perform optimizations on this type of allocation.
+    DedicatedImage(vk::Image),
+    /// The memory for this resource will be allocated and managed by gpu-allocator.
+    GpuAllocatorManaged,
+}
 
 #[derive(Clone, Debug)]
 pub struct AllocationCreateDesc<'a> {
@@ -23,16 +39,116 @@ pub struct AllocationCreateDesc<'a> {
     pub location: MemoryLocation,
     /// If the resource is linear (buffer / linear texture) or a regular (tiled) texture.
     pub linear: bool,
+    /// Determines how this allocation should be managed.
+    pub allocation_scheme: AllocationScheme,
 }
 
+/// Wrapper type to only mark a raw pointer [`Send`] + [`Sync`] without having to
+/// mark the entire [`Allocation`] as such, instead relying on the compiler to
+/// auto-implement this or fail if fields are added that violate this constraint
+#[derive(Clone, Copy, Debug)]
+pub(crate) struct SendSyncPtr(std::ptr::NonNull<std::ffi::c_void>);
+// Sending is fine because mapped_ptr does not change based on the thread we are in
+unsafe impl Send for SendSyncPtr {}
+// Sync is also okay because Sending &Allocation is safe: a mutable reference
+// to the data in mapped_ptr is never exposed while `self` is immutably borrowed.
+// In order to break safety guarantees, the user needs to `unsafe`ly dereference
+// `mapped_ptr` themselves.
+unsafe impl Sync for SendSyncPtr {}
+
 pub struct AllocatorCreateDesc {
     pub instance: ash::Instance,
     pub device: ash::Device,
     pub physical_device: ash::vk::PhysicalDevice,
     pub debug_settings: AllocatorDebugSettings,
     pub buffer_device_address: bool,
+    pub allocation_sizes: AllocationSizes,
 }
 
+/// A piece of allocated memory.
+///
+/// Could be contained in its own individual underlying memory object or as a sub-region
+/// of a larger allocation.
+///
+/// # Copying data into a CPU-mapped [`Allocation`]
+///
+/// You'll very likely want to copy data into CPU-mapped [`Allocation`]s in order to send that data to the GPU.
+/// Doing this data transfer correctly without invoking undefined behavior can be quite fraught and non-obvious<sup>[\[1\]]</sup>.
+///
+/// To help you do this correctly, [`Allocation`] implements [`presser::Slab`], which means you can directly
+/// pass it in to many of `presser`'s [helper functions] (for example, [`copy_from_slice_to_offset`]).
+///
+/// In most cases, this will work perfectly. However, note that if you try to use an [`Allocation`] as a
+/// [`Slab`] and it is not valid to do so (if it is not CPU-mapped or if its `size > isize::MAX`),
+/// you will cause a panic. If you aren't sure about these conditions, you may use [`Allocation::try_as_mapped_slab`].
+///
+/// ## Example
+///
+/// Say we've created an [`Allocation`] called `my_allocation`, which is CPU-mapped.
+/// ```ignore
+/// let mut my_allocation: Allocation = my_allocator.allocate(...)?;
+/// ```
+///
+/// And we want to fill it with some data in the form of a `my_gpu_data: Vec<MyGpuVector>`, defined as such:
+///
+/// ```ignore
+/// // note that this is size(12) but align(16), thus we have 4 padding bytes.
+/// // this would mean a `&[MyGpuVector]` is invalid to cast as a `&[u8]`, but
+/// // we can still use `presser` to copy it directly in a valid manner.
+/// #[repr(C, align(16))]
+/// #[derive(Clone, Copy)]
+/// struct MyGpuVertex {
+///     x: f32,
+///     y: f32,
+///     z: f32,
+/// }
+///
+/// let my_gpu_data: Vec<MyGpuData> = make_vertex_data();
+/// ```
+///
+/// Depending on how the data we're copying will be used, the vulkan device may have a minimum
+/// alignment requirement for that data:
+///
+/// ```ignore
+/// let min_gpu_align = my_vulkan_device_specifications.min_alignment_thing;
+/// ```
+///
+/// Finally, we can use [`presser::copy_from_slice_to_offset_with_align`] to perform the copy,
+/// simply passing `&mut my_allocation` since [`Allocation`] implements [`Slab`].
+///
+/// ```ignore
+/// let copy_record = presser::copy_from_slice_to_offset_with_align(
+///     &my_gpu_data[..], // a slice containing all elements of my_gpu_data
+///     &mut my_allocation, // our Allocation
+///     0, // start as close to the beginning of the allocation as possible
+///     min_gpu_align, // the minimum alignment we queried previously
+/// )?;
+/// ```
+///
+/// It's important to note that the data may not have actually been copied starting at the requested
+/// `start_offset` (0 in the example above) depending on the alignment of the underlying allocation
+/// as well as the alignment requirements of `MyGpuVector` and the `min_gpu_align` we passed in. Thus,
+/// we can query the `copy_record` for the actual starting offset:
+///
+/// ```ignore
+/// let actual_data_start_offset = copy_record.copy_start_offset;
+/// ```
+///
+/// ## Safety
+///
+/// It is technically not fully safe to use an [`Allocation`] as a [`presser::Slab`] because we can't validate that the
+/// GPU is not using the data in the buffer while `self` is borrowed. However, trying
+/// to validate this statically is really hard and the community has basically decided that
+/// requiring `unsafe` for functions like this creates too much "unsafe-noise", ultimately making it
+/// harder to debug more insidious unsafety that is unrelated to GPU-CPU sync issues.
+///
+/// So, as would always be the case, you must ensure the GPU
+/// is not using the data in `self` for the duration that you hold the returned [`MappedAllocationSlab`].
+///
+/// [`Slab`]: presser::Slab
+/// [`copy_from_slice_to_offset`]: presser::copy_from_slice_to_offset
+/// [helper functions]: presser#functions
+/// [\[1\]]: presser#motivation
 #[derive(Debug)]
 pub struct Allocation {
     chunk_id: Option<std::num::NonZeroU64>,
@@ -41,36 +157,75 @@ pub struct Allocation {
     memory_block_index: usize,
     memory_type_index: usize,
     device_memory: vk::DeviceMemory,
-    mapped_ptr: Option<std::ptr::NonNull<std::ffi::c_void>>,
-
+    mapped_ptr: Option<SendSyncPtr>,
+    dedicated_allocation: bool,
+    memory_properties: vk::MemoryPropertyFlags,
     name: Option<Box<str>>,
 }
 
-// Sending is fine because mapped_ptr does not change based on the thread we are in
-unsafe impl Send for Allocation {}
-// Sync is also okay because Sending &Allocation is safe: a mutable reference
-// to the data in mapped_ptr is never exposed while `self` is immutably borrowed.
-// In order to break safety guarantees, the user needs to `unsafe`ly dereference
-// `mapped_ptr` themselves.
-unsafe impl Sync for Allocation {}
-
 impl Allocation {
+    /// Tries to borrow the CPU-mapped memory that backs this allocation as a [`presser::Slab`], which you can then
+    /// use to safely copy data into the raw, potentially-uninitialized buffer.
+    /// See [the documentation of Allocation][Allocation#example] for an example of this.
+    ///
+    /// Returns [`None`] if `self.mapped_ptr()` is `None`, or if `self.size()` is greater than `isize::MAX` because
+    /// this could lead to undefined behavior.
+    ///
+    /// Note that [`Allocation`] implements [`Slab`] natively, so you can actually pass this allocation as a [`Slab`]
+    /// directly. However, if `self` is not actually a valid [`Slab`] (this function would return `None` as described above),
+    /// then trying to use it as a [`Slab`] will panic.
+    ///
+    /// # Safety
+    ///
+    /// See the note about safety in [the documentation of Allocation][Allocation#safety]
+    ///
+    /// [`Slab`]: presser::Slab
+    // best to be explicit where the lifetime is coming from since we're doing unsafe things
+    // and relying on an inferred liftime type in the PhantomData below
+    #[allow(clippy::needless_lifetimes)]
+    pub fn try_as_mapped_slab<'a>(&'a mut self) -> Option<MappedAllocationSlab<'a>> {
+        let mapped_ptr = self.mapped_ptr()?.cast().as_ptr();
+
+        if self.size > isize::MAX as _ {
+            return None;
+        }
+
+        // this will always succeed since size is <= isize::MAX which is < usize::MAX
+        let size = self.size as usize;
+
+        Some(MappedAllocationSlab {
+            _borrowed_alloc: PhantomData,
+            mapped_ptr,
+            size,
+        })
+    }
+
     pub fn chunk_id(&self) -> Option<std::num::NonZeroU64> {
         self.chunk_id
     }
 
+    ///Returns the [`vk::MemoryPropertyFlags`] of this allocation.
+    pub fn memory_properties(&self) -> vk::MemoryPropertyFlags {
+        self.memory_properties
+    }
+
     /// Returns the [`vk::DeviceMemory`] object that is backing this allocation.
     /// This memory object can be shared with multiple other allocations and shouldn't be freed (or allocated from)
     /// without this library, because that will lead to undefined behavior.
     ///
     /// # Safety
-    /// The result of this function can safely be used to pass into [`vk::DeviceFnV1_0::bind_buffer_memory()`],
-    /// [`vk::DeviceFnV1_0::bind_image_memory()`] etc. It is exposed for this reason. Keep in mind to also
+    /// The result of this function can safely be used to pass into [`ash::Device::bind_buffer_memory()`],
+    /// [`ash::Device::bind_image_memory()`] etc. It is exposed for this reason. Keep in mind to also
     /// pass [`Self::offset()`] along to those.
     pub unsafe fn memory(&self) -> vk::DeviceMemory {
         self.device_memory
     }
 
+    /// Returns [`true`] if this allocation is using a dedicated underlying allocation.
+    pub fn is_dedicated(&self) -> bool {
+        self.dedicated_allocation
+    }
+
     /// Returns the offset of the allocation on the [`vk::DeviceMemory`].
     /// When binding the memory to a buffer or image, this offset needs to be supplied as well.
     pub fn offset(&self) -> u64 {
@@ -85,7 +240,7 @@ impl Allocation {
     /// Returns a valid mapped pointer if the memory is host visible, otherwise it will return None.
     /// The pointer already points to the exact memory region of the suballocation, so no offset needs to be applied.
     pub fn mapped_ptr(&self) -> Option<std::ptr::NonNull<std::ffi::c_void>> {
-        self.mapped_ptr
+        self.mapped_ptr.map(|SendSyncPtr(p)| p)
     }
 
     /// Returns a valid mapped slice if the memory is host visible, otherwise it will return None.
@@ -119,17 +274,72 @@ impl Default for Allocation {
             memory_type_index: !0,
             device_memory: vk::DeviceMemory::null(),
             mapped_ptr: None,
+            memory_properties: vk::MemoryPropertyFlags::empty(),
             name: None,
+            dedicated_allocation: false,
         }
     }
 }
 
+/// A wrapper struct over a borrowed [`Allocation`] that infallibly implements [`presser::Slab`].
+///
+/// This type should be acquired by calling [`Allocation::try_as_mapped_slab`].
+pub struct MappedAllocationSlab<'a> {
+    _borrowed_alloc: PhantomData<&'a mut Allocation>,
+    mapped_ptr: *mut u8,
+    size: usize,
+}
+
+// SAFETY: See the safety comment of Allocation::as_mapped_slab above.
+unsafe impl<'a> presser::Slab for MappedAllocationSlab<'a> {
+    fn base_ptr(&self) -> *const u8 {
+        self.mapped_ptr
+    }
+
+    fn base_ptr_mut(&mut self) -> *mut u8 {
+        self.mapped_ptr
+    }
+
+    fn size(&self) -> usize {
+        self.size
+    }
+}
+
+// SAFETY: See the safety comment of Allocation::as_mapped_slab above.
+unsafe impl presser::Slab for Allocation {
+    fn base_ptr(&self) -> *const u8 {
+        self.mapped_ptr
+            .expect("tried to use a non-mapped Allocation as a Slab")
+            .0
+            .as_ptr()
+            .cast()
+    }
+
+    fn base_ptr_mut(&mut self) -> *mut u8 {
+        self.mapped_ptr
+            .expect("tried to use a non-mapped Allocation as a Slab")
+            .0
+            .as_ptr()
+            .cast()
+    }
+
+    fn size(&self) -> usize {
+        if self.size > isize::MAX as _ {
+            panic!("tried to use an Allocation with size > isize::MAX as a Slab")
+        }
+        // this will always work if the above passed
+        self.size as usize
+    }
+}
+
 #[derive(Debug)]
 pub(crate) struct MemoryBlock {
     pub(crate) device_memory: vk::DeviceMemory,
     pub(crate) size: u64,
-    pub(crate) mapped_ptr: *mut std::ffi::c_void,
+    pub(crate) mapped_ptr: Option<SendSyncPtr>,
     pub(crate) sub_allocator: Box<dyn allocator::SubAllocator>,
+    #[cfg(feature = "visualizer")]
+    pub(crate) dedicated_allocation: bool,
 }
 
 impl MemoryBlock {
@@ -138,8 +348,9 @@ impl MemoryBlock {
         size: u64,
         mem_type_index: usize,
         mapped: bool,
-        dedicated: bool,
         buffer_device_address: bool,
+        allocation_scheme: AllocationScheme,
+        requires_personal_block: bool,
     ) -> Result<Self> {
         let device_memory = {
             let alloc_info = vk::MemoryAllocateInfo::builder()
@@ -148,35 +359,62 @@ impl MemoryBlock {
 
             let allocation_flags = vk::MemoryAllocateFlags::DEVICE_ADDRESS;
             let mut flags_info = vk::MemoryAllocateFlagsInfo::builder().flags(allocation_flags);
-            // TODO(max): Test this based on if the device has this feature enabled or not
+            // TODO(manon): Test this based on if the device has this feature enabled or not
             let alloc_info = if buffer_device_address {
                 alloc_info.push_next(&mut flags_info)
             } else {
                 alloc_info
             };
 
-            unsafe { device.allocate_memory(&alloc_info, None) }
-                .map_err(|_| AllocationError::OutOfMemory)?
-        };
+            // Flag the memory as dedicated if required.
+            let mut dedicated_memory_info = vk::MemoryDedicatedAllocateInfo::builder();
+            let alloc_info = match allocation_scheme {
+                AllocationScheme::DedicatedBuffer(buffer) => {
+                    dedicated_memory_info = dedicated_memory_info.buffer(buffer);
+                    alloc_info.push_next(&mut dedicated_memory_info)
+                }
+                AllocationScheme::DedicatedImage(image) => {
+                    dedicated_memory_info = dedicated_memory_info.image(image);
+                    alloc_info.push_next(&mut dedicated_memory_info)
+                }
+                AllocationScheme::GpuAllocatorManaged => alloc_info,
+            };
 
-        let mapped_ptr = if mapped {
-            unsafe {
-                device.map_memory(
-                    device_memory,
-                    0,
-                    vk::WHOLE_SIZE,
-                    vk::MemoryMapFlags::empty(),
-                )
-            }
-            .map_err(|_| {
-                unsafe { device.free_memory(device_memory, None) };
-                AllocationError::FailedToMap
+            unsafe { device.allocate_memory(&alloc_info, None) }.map_err(|e| match e {
+                vk::Result::ERROR_OUT_OF_DEVICE_MEMORY => AllocationError::OutOfMemory,
+                e => AllocationError::Internal(format!(
+                    "Unexpected error in vkAllocateMemory: {:?}",
+                    e
+                )),
             })?
-        } else {
-            std::ptr::null_mut()
         };
 
-        let sub_allocator: Box<dyn allocator::SubAllocator> = if dedicated {
+        let mapped_ptr = mapped
+            .then(|| {
+                unsafe {
+                    device.map_memory(
+                        device_memory,
+                        0,
+                        vk::WHOLE_SIZE,
+                        vk::MemoryMapFlags::empty(),
+                    )
+                }
+                .map_err(|e| {
+                    unsafe { device.free_memory(device_memory, None) };
+                    AllocationError::FailedToMap(e.to_string())
+                })
+                .and_then(|p| {
+                    std::ptr::NonNull::new(p).map(SendSyncPtr).ok_or_else(|| {
+                        AllocationError::FailedToMap("Returned mapped pointer is null".to_owned())
+                    })
+                })
+            })
+            .transpose()?;
+
+        let sub_allocator: Box<dyn allocator::SubAllocator> = if allocation_scheme
+            != AllocationScheme::GpuAllocatorManaged
+            || requires_personal_block
+        {
             Box::new(allocator::DedicatedBlockAllocator::new(size))
         } else {
             Box::new(allocator::FreeListAllocator::new(size))
@@ -187,11 +425,13 @@ impl MemoryBlock {
             size,
             mapped_ptr,
             sub_allocator,
+            #[cfg(feature = "visualizer")]
+            dedicated_allocation: allocation_scheme != AllocationScheme::GpuAllocatorManaged,
         })
     }
 
     fn destroy(self, device: &ash::Device) {
-        if !self.mapped_ptr.is_null() {
+        if self.mapped_ptr.is_some() {
             unsafe { device.unmap_memory(self.device_memory) };
         }
 
@@ -199,11 +439,6 @@ impl MemoryBlock {
     }
 }
 
-// `mapped_ptr` is safe to send or share across threads because
-// it is never exposed publicly through [`MemoryBlock`].
-unsafe impl Send for MemoryBlock {}
-unsafe impl Sync for MemoryBlock {}
-
 #[derive(Debug)]
 pub(crate) struct MemoryType {
     pub(crate) memory_blocks: Vec<Option<MemoryBlock>>,
@@ -215,16 +450,14 @@ pub(crate) struct MemoryType {
     pub(crate) buffer_device_address: bool,
 }
 
-const DEFAULT_DEVICE_MEMBLOCK_SIZE: u64 = 256 * 1024 * 1024;
-const DEFAULT_HOST_MEMBLOCK_SIZE: u64 = 64 * 1024 * 1024;
-
 impl MemoryType {
     fn allocate(
         &mut self,
         device: &ash::Device,
-        desc: &AllocationCreateDesc,
+        desc: &AllocationCreateDesc<'_>,
         granularity: u64,
-        backtrace: Option<&str>,
+        backtrace: Option<backtrace::Backtrace>,
+        allocation_sizes: &AllocationSizes,
     ) -> Result<Allocation> {
         let allocation_type = if desc.linear {
             AllocationType::Linear
@@ -236,23 +469,27 @@ impl MemoryType {
             .memory_properties
             .contains(vk::MemoryPropertyFlags::HOST_VISIBLE)
         {
-            DEFAULT_HOST_MEMBLOCK_SIZE
+            allocation_sizes.host_memblock_size
         } else {
-            DEFAULT_DEVICE_MEMBLOCK_SIZE
+            allocation_sizes.device_memblock_size
         };
 
         let size = desc.requirements.size;
         let alignment = desc.requirements.alignment;
 
-        // Create a dedicated block for large memory allocations
-        if size > memblock_size {
+        let dedicated_allocation = desc.allocation_scheme != AllocationScheme::GpuAllocatorManaged;
+        let requires_personal_block = size > memblock_size;
+
+        // Create a dedicated block for large memory allocations or allocations that require dedicated memory allocations.
+        if dedicated_allocation || requires_personal_block {
             let mem_block = MemoryBlock::new(
                 device,
                 size,
                 self.memory_type_index,
                 self.mappable,
-                true,
                 self.buffer_device_address,
+                desc.allocation_scheme,
+                requires_personal_block,
             )?;
 
             let mut block_index = None;
@@ -292,10 +529,12 @@ impl MemoryType {
                 offset,
                 size,
                 memory_block_index: block_index,
-                memory_type_index: self.memory_type_index as usize,
+                memory_type_index: self.memory_type_index,
                 device_memory: mem_block.device_memory,
-                mapped_ptr: std::ptr::NonNull::new(mem_block.mapped_ptr),
+                mapped_ptr: mem_block.mapped_ptr,
+                memory_properties: self.memory_properties,
                 name: Some(desc.name.into()),
+                dedicated_allocation,
             });
         }
 
@@ -308,14 +547,15 @@ impl MemoryType {
                     allocation_type,
                     granularity,
                     desc.name,
-                    backtrace,
+                    backtrace.clone(),
                 );
 
                 match allocation {
                     Ok((offset, chunk_id)) => {
-                        let mapped_ptr = if !mem_block.mapped_ptr.is_null() {
-                            let offset_ptr = unsafe { mem_block.mapped_ptr.add(offset as usize) };
-                            std::ptr::NonNull::new(offset_ptr)
+                        let mapped_ptr = if let Some(SendSyncPtr(mapped_ptr)) = mem_block.mapped_ptr
+                        {
+                            let offset_ptr = unsafe { mapped_ptr.as_ptr().add(offset as usize) };
+                            std::ptr::NonNull::new(offset_ptr).map(SendSyncPtr)
                         } else {
                             None
                         };
@@ -324,9 +564,11 @@ impl MemoryType {
                             offset,
                             size,
                             memory_block_index: mem_block_i,
-                            memory_type_index: self.memory_type_index as usize,
+                            memory_type_index: self.memory_type_index,
                             device_memory: mem_block.device_memory,
+                            memory_properties: self.memory_properties,
                             mapped_ptr,
+                            dedicated_allocation: false,
                             name: Some(desc.name.into()),
                         });
                     }
@@ -335,7 +577,7 @@ impl MemoryType {
                         _ => return Err(err),              // Unhandled error, return.
                     },
                 }
-            } else if empty_block_index == None {
+            } else if empty_block_index.is_none() {
                 empty_block_index = Some(mem_block_i);
             }
         }
@@ -345,8 +587,9 @@ impl MemoryType {
             memblock_size,
             self.memory_type_index,
             self.mappable,
-            false,
             self.buffer_device_address,
+            desc.allocation_scheme,
+            false,
         )?;
 
         let new_block_index = if let Some(block_index) = empty_block_index {
@@ -383,9 +626,9 @@ impl MemoryType {
             },
         };
 
-        let mapped_ptr = if !mem_block.mapped_ptr.is_null() {
-            let offset_ptr = unsafe { mem_block.mapped_ptr.add(offset as usize) };
-            std::ptr::NonNull::new(offset_ptr)
+        let mapped_ptr = if let Some(SendSyncPtr(mapped_ptr)) = mem_block.mapped_ptr {
+            let offset_ptr = unsafe { mapped_ptr.as_ptr().add(offset as usize) };
+            std::ptr::NonNull::new(offset_ptr).map(SendSyncPtr)
         } else {
             None
         };
@@ -395,13 +638,16 @@ impl MemoryType {
             offset,
             size,
             memory_block_index: new_block_index,
-            memory_type_index: self.memory_type_index as usize,
+            memory_type_index: self.memory_type_index,
             device_memory: mem_block.device_memory,
             mapped_ptr,
+            memory_properties: self.memory_properties,
             name: Some(desc.name.into()),
+            dedicated_allocation: false,
         })
     }
 
+    #[allow(clippy::needless_pass_by_value)]
     fn free(&mut self, allocation: Allocation, device: &ash::Device) -> Result<()> {
         let block_idx = allocation.memory_block_index;
 
@@ -441,16 +687,52 @@ pub struct Allocator {
     device: ash::Device,
     pub(crate) buffer_image_granularity: u64,
     pub(crate) debug_settings: AllocatorDebugSettings,
+    allocation_sizes: AllocationSizes,
 }
 
 impl fmt::Debug for Allocator {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        f.debug_struct("Allocator")
-            .field("memory_types", &self.memory_types)
-            .field("memory_heaps", &self.memory_heaps)
-            .field("buffer_image_granularity", &self.buffer_image_granularity)
-            .field("debug_settings", &self.debug_settings)
-            .finish()
+        let mut allocation_report = vec![];
+        let mut total_reserved_size_in_bytes = 0;
+
+        for memory_type in &self.memory_types {
+            for block in memory_type.memory_blocks.iter().flatten() {
+                total_reserved_size_in_bytes += block.size;
+                allocation_report.extend(block.sub_allocator.report_allocations())
+            }
+        }
+
+        let total_used_size_in_bytes = allocation_report.iter().map(|report| report.size).sum();
+
+        allocation_report.sort_by_key(|alloc| std::cmp::Reverse(alloc.size));
+
+        writeln!(
+            f,
+            "================================================================",
+        )?;
+        writeln!(
+            f,
+            "ALLOCATION BREAKDOWN ({} / {})",
+            fmt_bytes(total_used_size_in_bytes),
+            fmt_bytes(total_reserved_size_in_bytes),
+        )?;
+
+        let max_num_allocations_to_print = f.precision().map_or(usize::MAX, |n| n);
+        for (idx, alloc) in allocation_report.iter().enumerate() {
+            if idx >= max_num_allocations_to_print {
+                break;
+            }
+
+            writeln!(
+                f,
+                "{:max_len$.max_len$}\t- {}",
+                alloc.name,
+                fmt_bytes(alloc.size),
+                max_len = allocator::VISUALIZER_TABLE_MAX_ENTRY_NAME_LEN,
+            )?;
+        }
+
+        Ok(())
     }
 }
 
@@ -493,19 +775,6 @@ impl Allocator {
             }
         }
 
-        // NOTE(max): Test if there is any HOST_VISIBLE memory that does _not_
-        //            have the HOST_COHERENT flag, in that case we want to panic,
-        //            as we want to do cool things that we do not yet support
-        //            with that type of memory :)
-        let host_visible_not_coherent = memory_types.iter().any(|t| {
-            let flags = t.property_flags;
-            flags.contains(vk::MemoryPropertyFlags::HOST_VISIBLE)
-                && !flags.contains(vk::MemoryPropertyFlags::HOST_COHERENT)
-        });
-        if host_visible_not_coherent {
-            warn!("There is a memory type that is host visible, but not host coherent. It's time to upgrade our memory allocator to take advantage of this type of memory :)");
-        }
-
         let memory_types = memory_types
             .iter()
             .enumerate()
@@ -535,15 +804,16 @@ impl Allocator {
             device: desc.device.clone(),
             buffer_image_granularity: granularity,
             debug_settings: desc.debug_settings,
+            allocation_sizes: AllocationSizes::default(),
         })
     }
 
-    pub fn allocate(&mut self, desc: &AllocationCreateDesc) -> Result<Allocation> {
+    pub fn allocate(&mut self, desc: &AllocationCreateDesc<'_>) -> Result<Allocation> {
         let size = desc.requirements.size;
         let alignment = desc.requirements.alignment;
 
         let backtrace = if self.debug_settings.store_stack_traces {
-            Some(format!("{:?}", backtrace::Backtrace::new()))
+            Some(backtrace::Backtrace::new_unresolved())
         } else {
             None
         };
@@ -554,10 +824,8 @@ impl Allocator {
                 &desc.name, size, alignment
             );
             if self.debug_settings.log_stack_traces {
-                let backtrace = backtrace
-                    .clone()
-                    .unwrap_or(format!("{:?}", backtrace::Backtrace::new()));
-                debug!("Allocation stack trace: {}", &backtrace);
+                let backtrace = backtrace::Backtrace::new();
+                debug!("Allocation stack trace: {:?}", backtrace);
             }
         }
 
@@ -585,10 +853,7 @@ impl Allocator {
         if memory_type_index_opt.is_none() {
             let mem_loc_required_bits = match desc.location {
                 MemoryLocation::GpuOnly => vk::MemoryPropertyFlags::DEVICE_LOCAL,
-                MemoryLocation::CpuToGpu => {
-                    vk::MemoryPropertyFlags::HOST_VISIBLE | vk::MemoryPropertyFlags::HOST_COHERENT
-                }
-                MemoryLocation::GpuToCpu => {
+                MemoryLocation::CpuToGpu | MemoryLocation::GpuToCpu => {
                     vk::MemoryPropertyFlags::HOST_VISIBLE | vk::MemoryPropertyFlags::HOST_COHERENT
                 }
                 MemoryLocation::Unknown => vk::MemoryPropertyFlags::empty(),
@@ -612,7 +877,8 @@ impl Allocator {
                 &self.device,
                 desc,
                 self.buffer_image_granularity,
-                backtrace.as_deref(),
+                backtrace.clone(),
+                &self.allocation_sizes,
             )
         };
 
@@ -633,7 +899,8 @@ impl Allocator {
                     &self.device,
                     desc,
                     self.buffer_image_granularity,
-                    backtrace.as_deref(),
+                    backtrace,
+                    &self.allocation_sizes,
                 )
             } else {
                 allocation
